#include <errno.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#define serial_port "/dev/ttyS0"

struct Image {
  int     image_size;   /* # of bytes of image */
  char   *image;      /* image data */
  char    image_type[5];    /* extension of image (jpg, png, etc) */
  int     image_info_size;  /* # of entries in image_info */
  char  **image_info;   /* tag=value pairs */
};

static int initialized = 0;

static int Dimera_Initialise(void);
static int Dimera_Number_of_pics();
static struct Image * Dimera_Get_Picture(int picnum, int thumbnail);

void take(int which, int thumb) {
  char buf[32];
  int fd;
  struct Image *img;
  if (!initialized++) Dimera_Initialise();
  fprintf(stderr, "Retrieving %s %d", thumb?"thumbnail":"image", which);
  fflush(stderr);
  img = Dimera_Get_Picture(which, thumb);
  sprintf(buf, "dimera%s-%d.pgm", thumb?"-t":"", which);
  fprintf(stderr, " saving %s...", buf); fflush(stderr);
  fd = open(buf, O_WRONLY|O_CREAT, 0644);
  if (fd == -1) perror("open"); else {
    write(fd, img->image, img->image_size);
    close(fd);
    fprintf(stderr, " done.\n");
  }
  free(img->image);
  free(img);
}

void usage(char *progname) {
  printf("Usage:\n  %s <command> [<command> [...]]\n\n", progname);
  printf("Commands:\n");
  printf("  status         - show how many pictures are in the camera\n");
  printf("  thumb <N>      - download a thumbnail of picture N\n");
  printf("  get <N>        - download picture N\n");
  printf("  all            - download all pictures\n");
  printf("  allthumbs      - download a thumbnail for each picture\n");
  printf("  help           - this message\n\n");
  exit(0);
}

int main(int argc, char **argv) {
  int i;
  if (argc < 2) usage(argv[0]);
  for (i=1; i<argc; i++)
    if (!strcmp(argv[i], "status")) {
      int num;
      if (!initialized++) Dimera_Initialise();
      num = Dimera_Number_of_pics();
      printf("%d pictures\n", num);
    }
    else if (!strcmp(argv[i], "thumb"))
      take(atoi(argv[++i]), 1);
    else if (!strcmp(argv[i], "get"))
      take(atoi(argv[++i]), 0);
    else if (!strcmp(argv[i], "all")) {
      int i;
      int num;
      if (!initialized++) Dimera_Initialise();
      num = Dimera_Number_of_pics();
      printf("%d pictures\n", num);
      for (i=0; i<num; i++)
        take(i, 0);
    }
    else if (!strcmp(argv[i], "allthumbs")) {
      int i;
      int num;
      if (!initialized++) Dimera_Initialise();
      num = Dimera_Number_of_pics();
      printf("%d pictures\n", num);
      for (i=0; i<num; i++)
        take(i, 1);
    }
    else if (!strcmp(argv[i], "help"))
      usage(argv[0]);
    else {
      printf("%s: invalid command: %s (try \"%s help\")\n",
        argv[0], argv[i], argv[0]);
      exit(-1);
    }
  return 0;
}


#define DIMERA_TRACE

#define DIMERA_CMD_ACK    0x21

#define  DIMERA_SETRATE    0x69
#define  DIMERA_AT1    0x41
#define  DIMERA_AT2    0x54
#define  DIMERA_AT3    0x0d
#define DIMERA_ENQ    0x05
#define DIMERA_STAT    0x55
#define DIMERA_THUMB    0x61
#define DIMERA_VIEW    0x21
#define DIMERA_VIEWBIG    0xFB  /* select large viewfinder image size */
#define DIMERA_IMAGE_INFO  0x71
#define DIMEAR_RETRIEVE_IMAGE  0x59
#define DIMERA_GET_LINES  0x49
#define DIMERA_GET_LINES_SZ  14  /* size of get lines cmd */

static unsigned char  Dimera_get_lines_std[ DIMERA_GET_LINES_SZ ] = {
  DIMERA_GET_LINES, 0x00, 0x00, 0x1a, 0x00, 0x04,
  0x00, 0x50, 0x00, 0x10, 0x01, 0x80, 0x00, 0x00 };

static unsigned char  Dimera_get_lines_fine[ DIMERA_GET_LINES_SZ ] = {
  DIMERA_GET_LINES, 0x00, 0x00, 0x1e, 0x00, 0x04,
  0x00, 0xa0, 0x00, 0x20, 0x01, 0x80, 0x00, 0x00 };

static unsigned char  imagebuffer[307200];

/* image sizes */
#define THUMBNAIL_SZ    3840  /* 64x48 */

static char  Dimera_thumbhdr[] =
"P5\n# Dimera 3500 Thumbnail\n64 48\n255\n";

#define STDIMAGE_SZ    76800  /* 320 x 240 */
#define FINEIMAGE_SZ    307200  /* 640 x 480 */

static char  Dimera_finehdr[] =
"P5\n# Dimera 3500 Image\n640 480\n255\n";

static char  Dimera_stdhdr[] =
"P5\n# Dimera 3500 Image\n320 240\n255\n";

static int  cam_fd = -1;
static char  msg_buf[64];

static unsigned char  enq[3];

#define ERROR(s) { \
  fprintf( stderr, "%s: %s\n", __FUNCTION__, s ); }

#define update_status(s) { \
  fprintf(stderr, "Status: %s\n", s); }

#ifdef DIMERA_TRACE
#define TRACE( s ) { \
printf( "%s: ", __FUNCTION__); printf s ; }
#else
#define TRACE( s )
#endif

#ifdef DIMERA_DUMP
#define DUMP(b,n) dim_dump( __FUNCTION__, b, n)
#else
#define DUMP(b,n)
#endif

static unsigned char
sum( unsigned char *b, unsigned char c, int s)
{
  while ( s-- > 0 )
    c += *b++;
  return c;
}

unsigned char *
setrgb( unsigned char *b, long p )
{
  b[0] = p>>16;
  b[1] = p>>8;
  b[2] = p;
  return b + 3;
}

static void
mkppm( unsigned char *ppm, unsigned char *cam, int w, int h )
{
  memcpy( ppm, cam, w * h );
}

/*
 * return the difference in tenths of seconds, between t1 and t2.
 */
static long
timediff( struct timeval *t1, struct timeval *t2 )
{
  long  t;

  t = (t1->tv_sec - t2->tv_sec) * 10;
  t += (t1->tv_usec - t2->tv_usec) / 100000;
  return t;
}

static int
Dimera_read( unsigned char *b, int s, int timeout2, int timeout1 )
{
  int    n = 0;
  int    r, t;
  int    i = 0;
  struct timeval  start, now;
  int last = -10000;

  t = timeout1 ? timeout1 : timeout2;  /* set first byte timeout */
  gettimeofday( &start, NULL );

  do
  {
      /* limit reads to 1k segments */
    r = read( cam_fd, &b[n], s>1024?1024:s );
    if ( r > 0 )
    {
      n += r;
      s -= r;
      if (n - last >= 256) {
        last = n;
      }
      gettimeofday( &start, NULL );
      t = timeout2;
    } else if ( r < 0 ) {
      if ( errno != EAGAIN )
      {
        ERROR("Read error");
        return -1;
      }
      i++;
    }
    gettimeofday( &now, NULL );
  } while ( s > 0 && t > timediff( &now, &start ) );
  DUMP( b, n );
  if (n > 1000) { fprintf(stderr, "."); fflush(stderr); }
  return n;
}

static int
Dimera_send_command( unsigned char *cmd, int n, int ackTimeout )
{
  char  c;

  DUMP( cmd, n );
  if ( write( cam_fd, cmd, n ) != n )
    return 0;

  if ( Dimera_read( &c, 1, ackTimeout, 0 ) != 1 )
  {
    TRACE( ("ACK timeout cmd %02x\n", *cmd) );
    return 0;
  }
  if ( c != DIMERA_CMD_ACK )
  {
    TRACE( ("BAD ACK cmd %02x\n", *cmd) );
    return -1;
  }
  return 1;
}

static int
Dimera_set_rate( int rate )
{
  unsigned char  cmd[2];
  struct termios  tio;
  int    speed;

  cmd[0] = DIMERA_SETRATE;
  switch ( rate )
  {
  case 9600:
    cmd[1] = 0;
    speed = B9600;
    break;

  case 19200:
    cmd[1] = 2;
    speed = B19200;
    break;

  case 38400:
    cmd[1] = 4;
    speed = B38400;
    break;

  case 57600:
    cmd[1] = 5;
    speed = B57600;
    break;

  case 115200:
    cmd[1] = 7;
    speed = B115200;
    break;

  default:
    ERROR("Invalid baud rate");
    return 0;

  }

  if ( Dimera_send_command( cmd, 2, 20 ) < 1 )
  {
    ERROR("Command failure: SetRate");
    return 0;
  }

  if (tcgetattr( cam_fd, &tio ) < 0 )
  {
    ERROR("Can't get serial port info");
    return 0;
  }
  cfsetospeed( &tio, speed );
  cfsetispeed( &tio, speed );
  if (tcsetattr( cam_fd, TCSANOW, &tio ) < 0 )
  {
    ERROR("Can't set serial port info");
    return 0;
  }
  snprintf( msg_buf, sizeof( msg_buf ), "Port Speed %d", rate );
  return 1;
}

static int
Dimera_port_open()
{
  struct termios  tio;
  unsigned char  b[3];
  int    i;

  if ( cam_fd == -1 )
    close( cam_fd );

#if 0
  cam_fd = open( serial_port, O_RDWR|O_NONBLOCK );
#else
  cam_fd = open( serial_port, O_RDWR);
#endif

  snprintf( msg_buf, sizeof( msg_buf ), "%s opened", serial_port );
  update_status( msg_buf );

  TRACE( ("%s opened\n", serial_port) );

  tcgetattr( cam_fd, &tio );
  cfmakeraw( &tio );
  tio.c_cc[VMIN] = 0;
  tio.c_cc[VTIME] = 1;

  cfsetospeed( &tio, B115200 );
  cfsetispeed( &tio, B115200 );
  if ( tcsetattr( cam_fd, TCSANOW, &tio ) < 0 )
    return 0;
  
  if ( tcsendbreak( cam_fd, 0 ) < 0 )
    return 0;
  TRACE( ("send break\n") );

  /*Dimera_set_rate( 38400 );*/
  Dimera_set_rate(115200);

  /*
   * the following commands, are sent because I know it works
   * I do not know if it is necessary
   */
  b[0] = DIMERA_AT1;
  b[1] = DIMERA_AT2;
  b[2] = DIMERA_AT3;
  Dimera_send_command( b, 3, 20 );

  b[0] = DIMERA_ENQ;
  Dimera_send_command( b, 1, 20 );

  if ( (i = Dimera_read( enq, 3, 10, 0 )) != 3 )
  {
    if ( i < 0 )
    {
      snprintf( msg_buf, sizeof( msg_buf ),
        "Port Open Initialiation Failed %s",
        strerror( errno ) );
      ERROR( msg_buf );
      update_status( msg_buf );
      return 0;
    } else {
      snprintf( msg_buf, sizeof( msg_buf ),
        "Port Open Initialiation Short Read" );
      ERROR( msg_buf );
      update_status( msg_buf );
    }
  }
  if ( enq[0] != 0x1d || enq[1] != 06 || enq[2] != 0x62 )
  {
    snprintf( msg_buf, sizeof( msg_buf ), "ENQ %02x %02x %02x",
        enq[0], enq[1], enq[2] );
    ERROR( msg_buf );
    update_status( msg_buf );
  }

  return 1;
}

static int
Dimera_Initialise(void) 
{


TRACE( ("Initialise\n") );
  if ( Dimera_port_open() == 0 )
  {
    ERROR("Couldn't open camera device");
    return 0;
  }

  return 1;
}

static int
Dimera_Number_of_pics() 
{
  int    pictures, n;
  unsigned char  b[2];

  b[0] = DIMERA_STAT;
  if ( (n = Dimera_send_command( b, 1, 10 )) == 1)
  {
    if ( (n = Dimera_read( b, 2, 10, 0 )) != 2 )
    {
      ERROR( "Status request failed" );
    } else {
      pictures = b[0]+(b[1]<<8);
      return pictures;
    }
  }

  return 0;
}

static void
Dimera_Get_Thumbnail( int picnum, struct Image *image )
{
  unsigned char  b[3];
  int    n;
  unsigned char  c = 0;

  if ( !(image->image = (unsigned char *) malloc( THUMBNAIL_SZ +
      sizeof( Dimera_thumbhdr ) - 1 )) )
  {
    ERROR( "Get Thumbnail, allocation failed" );
    image->image_size = 0;
    return;
  }

    /* set image size */
  image->image_size = THUMBNAIL_SZ + sizeof( Dimera_thumbhdr ) - 1;

    /* set image header */
  memcpy( image->image, Dimera_thumbhdr, sizeof( Dimera_thumbhdr ) - 1 );

    /* set image type */
  strcpy( image->image_type, "pgm" );

  image->image_info_size = 0;
  b[0] = DIMERA_THUMB;
  b[1] = picnum;
  b[2] = 0;

    /* send the Get Thumbnail command */
  if ( (n = Dimera_send_command( b, 3, 10 )) != 1)
  {
    ERROR( "Get Thumbnail, xmit failure" );
    return;
  }

  if ( (n = Dimera_read( b, 3, 10, 0 )) == 3 )
  {
    c = sum( b, c, 3 );
    if ( (n = Dimera_read( image->image +
        sizeof( Dimera_thumbhdr ) - 1, THUMBNAIL_SZ,
        10, 0 )) == THUMBNAIL_SZ )
    {
      c = sum( image->image + sizeof( Dimera_thumbhdr ) - 1,
        c, THUMBNAIL_SZ );
          /* read check digit */
      if ( Dimera_read( b, 3, 10, 0 ) != 1 )
      {
        ERROR( "Get Thumbnail, read size" );
        free( image->image );
        image->image_size = 0;
        return;
      }
      if ( c != b[0] )
      {
        ERROR( "Get Thumbnail, checksum mismatch" );
      }
      return;
    }
  }

  ERROR( "Get Thumbnail, xmit failure" );
  free( image->image );
  image->image_size = 0;
  return;
}


static void
Dimera_Get_Full_Image( int picnum, struct Image *image )
{
  unsigned char  b[14], *p, *template, *hdr, c = 0;
  int    n, i;
  int    lines;    /* lines per read */
  int    readsize;  /* bytes per read */
  int    height;
  int    width;

  image->image_size = 0;

  b[0] = DIMERA_IMAGE_INFO;
  b[1] = picnum;
  b[2] = 0;
    /* send the Get Image Info command */
  if ( (n = Dimera_send_command( b, 3, 10 )) != 1)
  {
    ERROR( "Get Image, xmit failure" );
    return;
  }

  if ( (n = Dimera_read( b, 3, 10, 0 )) != 3 )
  {
    
    ERROR( "Get Full Image, getinfo failed" );
    return;
  }

    /* get image size */
  if ( b[2] & 0x80 )
  {
    image->image_size = STDIMAGE_SZ + sizeof( Dimera_stdhdr ) - 1;
    width = 320;
    height = 240;
    lines = 16;
    readsize = lines * width;
    template = Dimera_get_lines_std;
    hdr = Dimera_stdhdr;
  } else {
    image->image_size = FINEIMAGE_SZ + sizeof( Dimera_finehdr ) - 1;
    width = 640;
    height = 480;
    lines = 32;
    readsize = lines * width;
    template = Dimera_get_lines_fine;
    hdr = Dimera_finehdr;
  }

    /* set image type */
  strcpy( image->image_type, "pnm" );

  image->image_info_size = 0;

  b[0] = DIMEAR_RETRIEVE_IMAGE;
  b[1] = picnum;
  b[2] = 0;
    /* send the Retrieve Image command */
  if ( (n = Dimera_send_command( b, 3, 500 )) != 1)
  {
    ERROR( "Get Image, xmit failure" );
    image->image_size = 0;
    return;
  }

  for ( i = 4, p = imagebuffer; i < height+4 ; i += lines )
  {
    memcpy( b, template, DIMERA_GET_LINES_SZ );
    b[1] = i;
    b[2] = i>>8;
    if ((n = Dimera_send_command( b, DIMERA_GET_LINES_SZ, 10)) !=
        1 )
    {
      ERROR( "Get Image, xmit failure" );
      image->image_size = 0;
      return;
    }
    if ( (n = Dimera_read( p, readsize, 10, 0 )) != readsize )
    {
      ERROR( "Get Image, read failure" );
      image->image_size = 0;
      return;
    }
    c = sum( p, 0, readsize );

      /* read check digit */
    if (Dimera_read( b, 3, 10, 0 ) != 1 )
    {
      ERROR( "Get Image, read size" );
      image->image_size = 0;
      return;
    }
    if ( b[0] != c )
    {
      ERROR( "Get Image, bad checksum" );
      image->image_size = 0;
      return;
    }
    p += readsize;
  }

  if ( (image->image = (unsigned char *) malloc( image->image_size )) ==
      NULL )
  {
    image->image_size = 0;
    return;
  }

  memcpy( image->image, hdr, sizeof( Dimera_stdhdr ) );

  DUMP( image->image, 64 );

  mkppm( image->image + (sizeof( Dimera_stdhdr) - 1), imagebuffer,
      width, height );
  DUMP( image->image, 320 );

  return;
}

static struct Image *
Dimera_Get_Picture(int picnum, int thumbnail)
{
  struct Image  *image = (struct Image *)malloc( sizeof( struct Image));

  if ( image == NULL )
  {
    ERROR( "Get Picture, allocation failed" );
  } else {

    if ( thumbnail )
    {
      Dimera_Get_Thumbnail( picnum, image );
    } else {
      Dimera_Get_Full_Image( picnum, image );
    }
    
  }
  if ( image->image_size )
    return image;
  free( image );
  return 0;
}
