#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libppm.h"

static char *getchunk(FILE *f);

int ppminit(struct ppm_image *p) {
  if (!p) return -1;
  p->data = NULL;
  p->width = p->height = 0;
  return 0;
}

int ppmalloc(struct ppm_image *p, int width, int height) {
  if (!p) return -1;
  if (p->data) free(p->data);
  if ((p->data = (PEL*)malloc(width*height*3)) == NULL) {
    p->width = p->height = 0;
    return -1;
  }
  p->width = width;
  p->height = height;
  return 0;
}
  
int ppmdestroy(struct ppm_image *p) {
  if (!p || !p->data) return -1;
  free(p->data);
  p->data = NULL;
  return 0;
}

int ppmread(FILE *infp, struct ppm_image *p) {
  char *buf;
  int ign, n, bytes;
  int greyscale = 0;

  if (!infp || !p) return -1;
  if (p->data) ppmdestroy(p);

  buf = getchunk(infp);
  if (!strcmp("P5", buf))
    greyscale = 1;
  else if (strcmp("P6", buf)) {
    fprintf(stderr, "Input does not appear to be in PPM format\n");
    return -1;
  }
  buf = getchunk(infp);   p->width  = atoi(buf);
  buf = getchunk(infp);   p->height = atoi(buf);
  buf = getchunk(infp);   ign       = atoi(buf);
  if (ign != 255) {
    fprintf(stderr, "Input does not appear to be in PPM format\n");
    return -1;
  }
  p->data = malloc(p->width*p->height*3);
  for (bytes=0; bytes<p->width*p->height*(greyscale ? 1 : 3); bytes+=n) {
    n=fread(p->data+bytes, 1, p->width*p->height*(greyscale ? 1 : 3)-bytes,
      infp);
    if (n <= 0) return -1;
  }
  if (greyscale)
    for (n=p->width*p->height-1; n>=0; n--)
      p->data[n*3+2] = p->data[n*3+1] = p->data[n*3] = p->data[n];

  return 0;
}

int ppmwrite(FILE *outfp, struct ppm_image *img) {
  if (!outfp || !img || !img->data || img->width <= 0 || img->height <= 0) return -1;
  fprintf(outfp, "P6\n%d %d 255\n", img->width, img->height);
  if (fwrite(img->data, img->width*img->height*3, 1, outfp) <= 0) return -1;
  return 0;
}

int ppmscale(struct ppm_image *p, int nw, int nh) {
  PEL *nimg;
  int x, y;

  if (!p || !p->data || nw <= 0 || nh <= 0) return -1;

  nimg = (unsigned char *)malloc(nw * nh * 3);
  if (!nimg) return -1;

  for (y=0; y<nh; y++)
    for (x=0; x<nw; x++) {
      int tmp1, tmp2;
      tmp1 = (x + y*nw) * 3;
      tmp2 = (x*p->width/nw + y*p->height/nh*p->width) * 3;
      nimg[tmp1] =     p->data[tmp2];
      nimg[tmp1 + 1] = p->data[tmp2 + 1];
      nimg[tmp1 + 2] = p->data[tmp2 + 2];
    }

  ppmdestroy(p);
  p->width = nw;
  p->height = nh;
  p->data = nimg;
  return 0;
}

int ppmcrop(struct ppm_image *p, int nw, int nh, PEL bg_r, PEL bg_g, PEL bg_b) {
  PEL *nimg, *exp;
  int x, y;
  int uh, uw;

  if (!p || !p->data || nw <= 0 || nh <= 0) return -1;

  nimg = (unsigned char *)malloc(nw * nh * 3);
  if (!nimg) return -1;
  uh = (nh < p->height) ? nh : p->height;
  uw = (nw < p->width) ?  nw : p->width;

  for (y=0; y<uh; y++) {
    memcpy(&nimg[y*nw*3], &p->data[y*p->width*3], uw*3);
    exp = &nimg[y*nw*3+uw*3];
    for (x=0; x<nw-uw; x++) {
      *exp++ = bg_r;
      *exp++ = bg_g;
      *exp++ = bg_b;
    }
  }
  exp = &nimg[uh*nw*3];
  for (x=0; x<(nh-uh)*nw; x++) {
    *exp++ = bg_r;
    *exp++ = bg_g;
    *exp++ = bg_b;
  }

  ppmdestroy(p);
  p->width = nw;
  p->height = nh;
  p->data = nimg;
  return 0;
}

int ppmgreyscale(struct ppm_image *p) {
  int i;
  PEL *q = p->data;
  if (!p || !p->data) return -1;
  for (i=0; i<p->width*p->height*3; i++,q+=3)
    *q = *(q+1) = *(q+2) = (*q + *(q+1) + *(q+2))/3;
  return 0;
}

int ppmdither2(struct ppm_image *p) {
  int i, dither_right, *dither_down;
  PEL *q = p->data;
  if (!p || !p->data) return -1;
  dither_down = (int*)malloc(sizeof(int)*p->width);
  if (!dither_down) return -1;
  bzero(dither_down, sizeof(int)*p->width);

  for (i=0; i<p->width*p->height; i++,q+=3) {
    int x = i%p->width, sum;
    if (!x) dither_right=0;
    sum = dither_right + dither_down[x] + *q + *(q+1) + *(q+2);
    if (sum <= 381) {
      *q = *(q+1) = *(q+2) = 0;
      dither_right = dither_down[x] = sum/2;
    }
    else {
      *q = *(q+1) = *(q+2) = 255;
      dither_right = dither_down[x] = (sum-765)/2;
    }
  }
  return 0;
}

static char *getchunk(FILE *f) {
  static char buf[200] = "\0";
  static char *p = buf;
  char *q, *tmp;
  if (!*p) {
    do
      fgets(buf, 200, f);
    while (*buf == '#');
    p = buf;
  }
  while (*p == ' ' || *p == '\t') p++;
  q = p;
  while (*q != ' ' && *q != '\t' && *q != '\n' && *q != '\0') q++;
  tmp = p;
  if (*q)     
    p = q+1;
  else
    p = q;
  *q = '\0';
  return tmp;
}
